--We will keep track of every smart loader
--each loader or unloader will be stored in global.ents table
--each entity of the table will contain either loader field, or a set of unloader-related fields.
--either {loader=entity} or {unloader=entity, inserter=entity}
--loader is simple inserter entity, unloader is entity built over any inserter with filter fields in it
----when each inserter is first placed under the unloader module it is checked
----if it has filter slots it is suitable
----names of suitable inserters along with their filter_count are stored in global.filters table
----names of all checked inserters are stored in global.checked table
--for loader:
----if it's imposible to put entity into it's target, we'll rotate it back to source
------the rotated entity is stored in another table global.rotate as {tick,entity}
----a few ticks later we'll return it's orientation to normal and remove it from global.rotate
----we won't check if inserter is placing items onto empty ground
----we won't check whether it succeeded in putting item in rotated position
--for unloader:
----we check whether inserter is valid, if not, we search for a new one
----we c`heck whether unloader filters are set
------if they are, we check for presence of a target of inserter
--------if the traget is present, we obtain item counts of items set in the unloader conditions
----------those items which are below the threshold are added to filter of inserter  
require "defines"
require "util"
interval=2*60

remote.add_interface("SmrtLdr",
{migrate2 = function()
--this function should transfer the data state from version 0.1.0 to 0.2.0
    for _,loader in pairs(global.loaders) do
        table.insert(global.ents,{loader=loader})
    end
end,
flush_filters=function()
--this forces recheck of every inserter type for unloader
--use if filter_count of already present entity has changed
    global.tested={}
    global.filters={}
    for _,it in pairs(global.ents) do
        it.inserter=nil
    end
end,
})
function init() 
    global.ents=global.ents or {}; 
    global.next_check=global.next_check or game.tick+interval;
    global.next_rotation=global.next_rotation or 0
    global.initial=global.initial or 1
    global.rotate=global.rotate or {}
    global.filters=global.filters or {}
    global.tested=global.tested or {}
end
script.on_init(function() init() end)
--game.on_load(function() init() end)

script.on_event({defines.events.on_built_entity,defines.events.on_robot_built_entity},
    function(event)
        if event.created_entity.name=='smart-loader' or event.created_entity.name=='smart-unloader' then
            if event.created_entity.name=='smart-loader' then
                table.insert(global.ents,{loader=event.created_entity})
            elseif event.created_entity.name=='smart-unloader' then
                table.insert(global.ents,{unloader=event.created_entity})
            end
            if #global.ents==1 then 
                global.next_check=game.tick+interval
                global.initial=1;
            end
        end
    end)
function loader_check(i,tick)
    local loader=global.ents[i].loader;
    if loader.held_stack.valid_for_read and 
        (not loader.drop_target or not loader.drop_target.can_insert(loader.held_stack)) then
        loader.direction=util.oppositedirection(loader.direction)
        table.insert(global.rotate,{tick=tick+20,entity=loader})
        global.next_rotation=global.rotate[1].tick
    end
end
function test_ins(ins)
    function try_set(ins,n)
        ins.set_filter('iron-plate',n)
    end
    if pcall(function () try_set(ins,1) end) then
        ins.clear_filter(1);
        global.filters[ins.name]=1
        while pcall(function () try_set(ins,global.filters[ins.name]+1) end) do
            global.filters[ins.name]=global.filters[ins.name]+1;
            ins.clear_filter(global.filters[ins.name]);
        end
    end
    global.tested[ins.name]=true
end
function unloader_check(i)
    local ent=global.ents[i]
    local unloader=ent.unloader;
    local inserter=ent.inserter
    if (not inserter) or (not inserter.valid) then
        ent.inserter=nil
        for _,ins in pairs(unloader.surface.find_entities_filtered(
         {type='inserter',area={
            {unloader.position.x-0.1,unloader.position.y-0.1,},
            {unloader.position.x+0.1,unloader.position.y+0.1,},
            }
         })) 
        do
            if ins.valid then
                if not global.tested[ins.name] then
                    test_ins(ins)
                end
                if global.filters[ins.name] then
                    ent.inserter=ins;
                    inserter=ins;
                    break;
                end
            end
        end
    end
    if inserter then
        local num=1
        for j=1,global.filters[inserter.name] do 
            inserter.clear_filter(j)
        end
        for _,param in pairs(unloader.get_circuit_condition(0).parameters) do
            if param.signal.name and param.signal.type=='item' then
                if inserter.drop_target then
                    if param.count>inserter.drop_target.get_item_count(param.signal.name) then
                        inserter.set_filter(param.signal.name,num)
                        inserter.direction=inserter.direction
                        num=num+1
                    end
                --else 
                    --inserter.set_filter(param.signal.name,num)
                    --num=num+1
                end
                if num>global.filters[inserter.name] then break end
            end
        end
    end
end
script.on_event(defines.events.on_tick, 
    function(event)
        if event.tick==global.next_rotation then
            for i=#global.rotate,1,-1 do
                local loader=global.rotate[i]
                if event.tick>=loader.tick then
                    loader.entity.direction=util.oppositedirection(loader.entity.direction)
                    table.remove(global.rotate,i)
                end
            end
            global.next_rotation=(global.rotate[1] and global.rotate[1].tick) or 0
        end
        if event.tick < global.next_check then return end
        local ents=global.ents
        for i=global.initial,#ents,interval do
            local ent=ents[i]
            if (ent.loader and not( ent.loader.valid)) or (ent.unloader and not (ent.unloader.valid)) then 
                table.remove(ents,i)
                return
            end
            if ent.loader then loader_check(i,event.tick)
            elseif ent.unloader then unloader_check(i)
            end
        end
        
        global.initial=(global.initial)%#ents+1;
        if #ents<interval then
            global.next_check=global.next_check+math.floor(interval/(#ents))
        end
    end)

function remov(entity)
    function get(ent)
        return (ent.loader or ent.unloader)
    end
    i=1;
    while i<=#global.ents and not(entity==get(global.ents[i])) do
        i=i+1;
    end
    if i<=#global.ents then 
        table.remove(global.ents,i); 
    end
end
script.on_event({defines.events.on_entity_died,defines.events.on_robot_pre_mined,defines.events.on_preplayer_mined_item,},
    function(event)
        if event.entity.name=='smart-loader' or event.entity.name=='smart-unloader' then
            remov(event.entity)
        end
    end)
